#include "framebuffer.h"
#include "math.h"
#include <iostream>
#include "scene.h"
#include "v3.h"
#include <iostream>

#include "tiffio.h"

using namespace std;

FrameBuffer::FrameBuffer(int u0, int v0,int _w, int _h) 
	: Fl_Gl_Window(u0, v0, _w, _h, 0) {

	w = _w;
	h = _h;
	pix = new unsigned int[w*h];

}


void FrameBuffer::draw() {

	glDrawPixels(w, h, GL_RGBA, GL_UNSIGNED_BYTE, pix);

}

int FrameBuffer::handle(int event) {

	switch (event)
	{
	case FL_KEYBOARD: {
		KeyboardHandle();
		return 0;
	}
	default:
		break;
	}
	return 0;
}

void FrameBuffer::KeyboardHandle() {

	int key = Fl::event_key();
	switch (key) {
	case ',': {
		cerr << "INFO: pressed ," << endl;
		break;
	}
	default:
		cerr << "INFO: do not understand keypress" << endl;
	}
}


void FrameBuffer::SetGuarded(int u, int v, unsigned int color) {

	if (u < 0 || u > w - 1 || v < 0 || v > h - 1)
		return;

	Set(u, v, color);

}


void FrameBuffer::Set(int u, int v, unsigned int color) {

	pix[(h - 1 - v)*w + u] = color;

}

void FrameBuffer::SetBGR(unsigned int bgr) {

	for (int uv = 0; uv < w*h; uv++)
		pix[uv] = bgr;

}


// load a tiff image to pixel buffer
void FrameBuffer::LoadTiff(char* fname) {
	TIFF* in = TIFFOpen(fname, "r");
	if (in == NULL) {
		cerr << fname << " could not be opened" << endl;
		return;
	}

	int width, height;
	TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &width);
	TIFFGetField(in, TIFFTAG_IMAGELENGTH, &height);
	if (w != width || h != height) {
		w = width;
		h = height;
		delete[] pix;
		pix = new unsigned int[w*h];
		size(w, h);
		glFlush();
		glFlush();
	}

	if (TIFFReadRGBAImage(in, w, h, pix, 0) == 0) {
		cerr << "failed to load " << fname << endl;
	}

	TIFFClose(in);
}

// save as tiff image
void FrameBuffer::SaveAsTiff(char *fname) {

	TIFF* out = TIFFOpen(fname, "w");

	if (out == NULL) {
		cerr << fname << " could not be opened" << endl;
		return;
	}

	TIFFSetField(out, TIFFTAG_IMAGEWIDTH, w);
	TIFFSetField(out, TIFFTAG_IMAGELENGTH, h);
	TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 4);
	TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);
	TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
	TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);

	for (uint32 row = 0; row < (unsigned int)h; row++) {
		TIFFWriteScanline(out, &pix[(h - row - 1) * w], row);
	}

	TIFFClose(out);
}


void FrameBuffer::DrawRectangle(int u0, int v0, int u1, int v1, unsigned int color) {

	if (!ClipToScreen(u0, v0, u1, v1))
		return;

	for (int u = u0; u <= u1; u++) {
		for (int v = v0; v <= v1; v++) {
			Set(u, v, color);
		}
	}

}

void FrameBuffer::DrawCircle(int u0, int v0, int r, unsigned int color) {

	int left = u0 - r;
	int right = u0 + r;
	int top = v0 - r;
	int bottom = v0 + r;
	if (!ClipToScreen(left, right, top, bottom))
		return;

	V3 cv(.5f + (float)u0, .5f + (float)v0, 0.0f);
	for (int u = left; u <= right; u++) {
		for (int v = top; v <= bottom; v++) {
			V3 pv(.5f + (float)u, .5f + (float)v, 0.0f);
			V3 dv = pv - cv;
			if (dv*dv > r*r)
				continue;
			Set(u, v, color);
		}
	}

}

int FrameBuffer::ClipToScreen(int &u0, int &v0, int &u1, int &v1) {


	if (u1 < 0)
		return 0;
	if (u0 > w - 1)
		return 0;
	if (v1 < 0)
		return 0;
	if (v0 > h - 1)
		return 0;

	if (u0 < 0)
		u0 = 0;
	if (v0 < 0)
		v0 = 0;
	if (u1 > w - 1)
		u1 = w - 1;
	if (v1 > h - 1)
		v1 = h - 1;
	return 1;

}

void FrameBuffer::DrawSegment(V3 p0, V3 c0, V3 p1, V3 c1) {

	V3 p1p0 = p1 - p0;
	V3 c1c0 = c1 - c0;
	int pixN;
	if (fabsf(p1p0[0]) > fabsf(p1p0[1])) {
		pixN = (int)(fabsf(p1p0[0]) + 1);
	}
	else {
		pixN = (int)(fabsf(p1p0[1]) + 1);
	}
	for (int stepsi = 0; stepsi < pixN; stepsi++) {
		float fracf = (float) stepsi / (float) (pixN - 1);
		V3 p = p0 + p1p0 * fracf;
		V3 c = c0 + c1c0 * fracf;
		int u = (int) p[0];
		int v = (int) p[1];
		SetGuarded(u, v, c.GetColor());
	}

}
